from uuid import uuid4


class PropertyOptionsDef:
    # children is list of type (PropertyOptionsDef | PropertyOptionsItemDef)
    def __init__(self, name: str, id: str = None, children=[]):
        self.name = name
        self.id = str(uuid4()) if id is None else id
        self.children = children

    def to_dict(self):
        return {
            "name": self.name,
            "id": self.id,
            "children": [child.to_dict() for child in self.children],
        }


class PropertyOptionsItemDef:
    def __init__(self, name: str, id: str = None):
        self.name = name
        self.id = str(uuid4()) if id is None else id

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
        }


class PropertyDef:
    # options is list of type (PropertyOptionsDef | PropertyOptionsItemDef)
    @staticmethod
    def multiple_choice_property(
        name: str,
        options: list,
        id: str = None,
        default_value: str = None,
        description: str = "",
        render_value: bool = False,
        is_required: bool = False,
        is_per_frame: bool = False,
    ):
        # options_list = [{"id": str(uuid4()), "name": item} for item in options]
        options_list = [_options.to_dict() for _options in options]
        id = str(uuid4()) if id is None else id
        return {
            "id": id,
            "name": name,
            "default_value": default_value,
            "description": description,
            "options": options_list,
            "render_value": render_value,
            "required": is_required,
            "per_frame": is_per_frame,
            "type": "radio",
        }

    # options is list of type (PropertyOptionsDef | PropertyOptionsItemDef)
    @staticmethod
    def multiple_selection_property(
        name: str,
        options: list,
        id: str = None,
        default_value: list = [],
        description: str = "",
        render_value: bool = False,
        is_required: bool = False,
        is_per_frame: bool = False,
    ):
        # options_list = [{"id": str(uuid4()), "name": item} for item in options]
        options_list = [_options.to_dict() for _options in options]
        id = str(uuid4()) if id is None else id
        return {
            "id": id,
            "name": name,
            "default_value": default_value,
            "description": description,
            "options": options_list,
            "render_value": render_value,
            "required": is_required,
            "per_frame": is_per_frame,
            "type": "checkbox",
        }

    @staticmethod
    def free_response_property(
        name: str,
        blank: bool = True,
        constraints: dict = {
            "digit": True,
            "space": True,
            "special": True,
            "alphabet": True,
        },
        id: str = None,
        default_value: str = "",
        description: str = "",
        render_value: bool = False,
        is_per_frame: bool = False,
    ):
        id = str(uuid4()) if id is None else id
        return {
            "id": id,
            "name": name,
            "default_value": default_value,
            "description": description,
            "constraints": constraints,
            "render_value": render_value,
            "blank": blank,
            "per_frame": is_per_frame,
            "type": "free response",
        }



from abc import ABC, abstractmethod
from .utils import unique_string_builder, get_name_from_unique_string


class LabelCreator(ABC):
    def __init__(self, object_classes_map):
        self.object_classes_map = object_classes_map
        self.label = None

    @abstractmethod
    def create(self):
        pass
    @abstractmethod
    def from_dict(self):
        pass

    def get_meta(self, class_name, class_type):
        return {
            "z_index": 1,
            "visible": True,
            "alpha": 1,
            "color": self.object_classes_map[unique_string_builder(
                class_name, class_type
            )]["color"],
        }
    
    def get_class_names(self):
        return [get_name_from_unique_string(x) for x in list(self.object_classes_map.keys())]
    
    def build(self):                
        if self.label is None:
            raise NotImplementedError("Label is not defined")

        if self.label.class_name is None:
            raise ValueError("class_name is not defined")

        if self.label.tracking_id is None:
            raise ValueError("tracking_id is not defined")

        return self.label.to_dict()



import random
from uuid import uuid4

from ..common.object_interface import ObjectInterfaceDef


class ObjectTrackingDef(ObjectInterfaceDef):
    def __init__(
        self,
        keypoints: list = [],
        object_groups: list = [],
        object_classes: list = [],
        annotation_types: list = [],
    ):
        super().__init__(
            keypoints=keypoints,
            object_groups=object_groups,
            object_classes=object_classes,
            annotation_types=annotation_types,
        )



from ..common.label import Label, ClassType
from ..common.utils import set_properties


class Cuboid(Label):
    def __init__(self, **attributes):
        super().__init__(attributes=attributes)
        self.frames = attributes.get("frames", [])
        self.annotation_type = ClassType.CUBOID

    def to_dict(self):
        return {
            "id": self.id,
            "class_name": self.class_name,
            "class_id": self.class_id,
            "annotation_type": self.annotation_type.value,
            "frames": [
                {
                    "num": frame["num"],
                    "properties": frame["properties"],
                    "annotation": frame["annotation"],
                }
                for frame in self.frames
            ],
            "properties": set_properties(self.properties_def, self.properties),
            "tracking_id": self.tracking_id,
        }



from .. import __version__
from ..video.categorization import CategorizationDef
from .object_tracking import ObjectTrackingDef


class LabelInterface:
    def __init__(
        self,
        categorization: CategorizationDef,
        object_tracking: ObjectTrackingDef,
        version: str = __version__,
    ):
        self._version = version
        self.type = "pointclouds-siesta"
        self.data_type = "pointclouds"
        self.categorization = categorization
        self.object_tracking = object_tracking

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, version):
        self._version = version

    @classmethod
    def get_default(cls):
        return cls(
            categorization=CategorizationDef.get_default(),
            object_tracking=ObjectTrackingDef.get_default(),
        )

    @classmethod
    def from_dict(cls, label_interface_dict: dict):
        return cls(
            categorization=CategorizationDef.from_dict(
                label_interface_dict["categorization"]
            ),
            object_tracking=ObjectTrackingDef.from_dict(
                label_interface_dict["object_tracking"]
            ),
            version=label_interface_dict["version"],
        )

    def set_categorization(self, categorization: CategorizationDef):
        self.categorization = categorization
        if not "image category" in self.object_tracking.annotation_types:
            self.object_tracking.annotation_types.append("image category")

    def set_object_tracking(self, object_tracking: ObjectTrackingDef):
        if "image category" in self.object_tracking.annotation_types:
            object_tracking.annotation_types.append("image category")
        self.object_tracking = object_tracking

    def to_dict(self):
        return {
            "type": self.type,
            "version": self.version,
            "data_type": self.data_type,
            "categorization": self.categorization.to_dict(),
            "object_tracking": self.object_tracking.to_dict(),
        }



from uuid import uuid4

from .. import __version__
from ..common.video_utils import (
    calculate_annotated_frame_count,
    calculate_video_properties_count,
)
from ..common.utils import unique_string_builder
from .label_creator import PointcloudsAnnotationBuilder, CuboidCreator
from ..exceptions import InvalidObjectException

class LabelInfo:
    @classmethod
    def _get_opt_map(cls, options):
        res = {}
        for opt in options:
            if "children" in opt:
                res.update(LabelInfo._get_opt_map(opt["children"]))
            else:
                res[opt["name"]] = opt
        return res

    @classmethod
    def _set_properties(cls, properties_def, properties):
        prop_def_map = {
            prop_def["name"]: prop_def for prop_def in properties_def
        }
        converted_properties = []
        for prop in properties:
            prop_def = prop_def_map[prop["name"]]
            if prop_def["type"] in ["radio", "dropdown", "checkbox"]:
                opt_map = LabelInfo._get_opt_map(prop_def["options"])
                if prop_def["type"] == "checkbox":
                    converted_properties.append(
                        {
                            "type": prop_def["type"],
                            "property_id": prop_def["id"],
                            "property_name": prop_def["name"],
                            "option_ids": [
                                opt_map[val]["id"] for val in prop["value"]
                            ],
                            "option_names": [
                                opt_map[val]["name"] for val in prop["value"]
                            ],
                        }
                    )
                else:
                    converted_properties.append(
                        {
                            "type": prop_def["type"],
                            "property_id": prop_def["id"],
                            "property_name": prop_def["name"],
                            "option_id": opt_map[prop["value"]]["id"],
                            "option_name": opt_map[prop["value"]]["name"],
                        }
                    )
            elif prop_def["type"] == "free response":
                converted_properties.append(
                    {
                        "type": prop_def["type"],
                        "property_id": prop_def["id"],
                        "property_name": prop_def["name"],
                        "value": prop["value"],
                    }
                )
        return converted_properties

    @classmethod
    def _get_properties(cls, properties_def, properties):
        prop_def_map = {
            prop_def["name"]: prop_def for prop_def in properties_def
        }
        converted_properties = []
        for prop in properties:
            prop_def = prop_def_map[prop["property_name"]]
            if prop_def["type"] in ["radio", "dropdown"]:
                converted_properties.append(
                    {
                        "name": prop["property_name"],
                        "value": prop["option_name"],
                    }
                )
            elif prop_def["type"] == "checkbox":
                converted_properties.append(
                    {
                        "name": prop["property_name"],
                        "value": prop["option_names"],
                    }
                )
            elif prop_def["type"] == "free response":
                converted_properties.append(
                    {
                        "name": prop["property_name"],
                        "value": prop["value"],
                    }
                )
        return converted_properties

    def __init__(self, label_interface, result=None):
        self.label_interface = label_interface
        self.object_classes_map = {
            unique_string_builder(
                object_class["name"],
                object_class["annotation_type"]
            ): object_class for object_class in label_interface.object_tracking.object_classes
        }
        if result is None:
            self.result = {}
            self.init_objects()
            self.init_categories()
        else:
            self.result = result

    def init_objects(self):
        self.result["objects"] = []

    def init_categories(self):
        self.result["categories"] = {"frames": [], "properties": []}

    def get_next_tracking_id(self):
        next_tracking_id = 0
        for obj in self.result["objects"]:
            if obj["tracking_id"] > next_tracking_id:
                next_tracking_id = obj["tracking_id"]
        next_tracking_id += 1

        return next_tracking_id

    def _get_property_info(self, properties_def, property_name):
        for prop_def in properties_def:
            if prop_def["name"] == property_name:
                return prop_def
        return None

    def _get_option_id(self, property_info, option_name):
        for opt in property_info["options"]:
            if opt["name"] == option_name:
                return opt["id"]
        return None

    # for example, if you want to add "Motion" property with "Walking" option,
    # you may give input property_dict as { "Motion": "Walking" } (radio type), { "Is carrying": ["Child", "Baggage"]} (checkbox type)
    def build_properties(self, class_name, annotation_type, property_dict, is_per_frame=False):
        result_properties = []

        unique_class_string = unique_string_builder(
            class_name,
            annotation_type
        )

        for property_name, option_value in property_dict.items():
            if option_value is None:
                continue
            
            property_info = self._get_property_info(
                self.object_classes_map[unique_class_string]["properties"],
                property_name
            )

            if property_info is None:
                continue
            if is_per_frame and not property_info["per_frame"]:
                continue
            if not is_per_frame and property_info["per_frame"]:
                continue

            if property_info["type"] in ["radio", "dropdown"]:
                option_id = self._get_option_id(
                    property_info,
                    option_value
                )
                if option_id is None:
                    continue
                
                result_properties.append({
                    "type": property_info["type"],
                    "property_id": property_info["id"],
                    "property_name": property_name,
                    "option_id": option_id,
                    "option_name": option_value
                })
            elif property_info["type"] == "checkbox":
                if not isinstance(option_value, list):
                    continue
                valid_option_names = []
                valid_option_ids = []
                for option_name in option_value:
                    option_id = self._get_option_id(
                        property_info,
                        option_name
                    )
                    if option_id is not None:
                        valid_option_names.append(option_name)
                        valid_option_ids.append(option_id)
                
                if len(valid_option_ids) == 0:
                    continue
                
                result_properties.append({
                    "type": property_info["type"],
                    "property_id": property_info["id"],
                    "property_name": property_name,
                    "option_id": valid_option_ids,
                    "option_name": valid_option_names
                })

            elif property_info["type"] == "free response":
                if not isinstance(option_value, str):
                    continue
                result_properties.append({
                    "type": property_info["type"],
                    "property_id": property_info["id"],
                    "property_name": property_name,
                    "value": option_value
                })
            else:
                continue

        return result_properties        

    def add_object(self, label):
        self.result["objects"].append(label)

    def remove_object(self, object_id):
        self.result["objects"] = [
            obj for obj in self.result["objects"] if obj["id"] != object_id
        ]

    def get_object_by_id(self, object_id):
        for obj in self.result["objects"]:
            if obj["id"] == object_id:
                return obj
        return None        

    def get_objects(self):
        ocm = self.object_classes_map
        try:
            if "objects" not in self.result:
                return []
            if not isinstance(self.result["objects"], list):
                return []
            return self.result["objects"]
        except Exception as e:
            return []

    def set_simple_categories(self, frames=None, properties=None):
        self.result["categories"] = {
            "frames": [
                {
                    "num": frame["num"],
                    "properties": LabelInfo._set_properties(
                        self.label_interface.categorization.properties,
                        frame["properties"],
                    ),
                }
                for frame in (frames if frames is not None else [])
            ],
            "properties": LabelInfo._set_properties(
                self.label_interface.categorization.properties,
                properties if properties is not None else [],
            ),
        }

    def get_simple_categories(self):
        try:
            simple_categories = {
                "frames": [
                    {
                        "num": frame["num"],
                        "properties": LabelInfo._get_properties(
                            self.label_interface.categorization.properties,
                            frame["properties"],
                        ),
                    }
                    for frame in self.result["categories"]["frames"]
                ],
                "properties": LabelInfo._get_properties(
                    self.label_interface.categorization.properties,
                    self.result["categories"]["properties"],
                ),
            }
            return simple_categories
        except:
            return {
                "frames": [],
                "properties": [],
            }

    def build_tags(self):
        classes_count = {}
        anno_count = {}
        classes_name = {}
        for obj in self.result["objects"]:
            cid = obj["class_id"]
            classes_count[cid] = classes_count.get(cid, 0) + 1
            anno_count[cid] = anno_count.get(cid, 0) + len(obj["frames"])
            classes_name[cid] = obj["class_name"]
        class_val = list(classes_name.values())

        categories_id = []
        if (
            "categories" in self.result
            and "properties" in self.result["categories"]
        ):
            properties_list = [self.result["categories"]["properties"]] + [
                frame["properties"]
                for frame in self.result["categories"]["frames"]
            ]
            for properties in properties_list:
                for prop in properties:
                    if "option_id" in prop:
                        categories_id.append(prop["option_id"])
                    elif "option_ids" in prop:
                        for o_id in prop["option_ids"]:
                            categories_id.append(o_id)
            class_val.extend(categories_id)

        return {
            "classes_id": list(classes_count.keys()),
            "class": class_val,
            "classes_count": [
                {
                    "id": k,
                    "name": classes_name[k],
                    "count": v,
                }
                for k, v in classes_count.items()
            ],
            "classes_properties_count": calculate_video_properties_count(
                self.result["objects"]
            ),
            "classes_annotation_count": [
                {
                    "id": k,
                    "name": classes_name[k],
                    "count": v,
                }
                for k, v in anno_count.items()
            ],
            "categories_id": categories_id,
            "annotated_frame_count": calculate_annotated_frame_count(
                self.result["objects"]
            ),
        }

    def build_info(self):
        ocm = self.object_classes_map
        return {
            "version": __version__,
            "meta": {
                "image_info": {},
                "edit_info": {
                    "objects": [
                        {
                            "id": obj["id"],
                            "color": ocm[unique_string_builder(
                                obj["class_name"],
                                obj["annotation_type"]
                            )]["color"],
                            "visible": True,
                            "selected": False,
                            "tracking_id": obj["tracking_id"],
                        }
                        for obj in self.result["objects"]
                    ]
                },
            },
            "result": self.result,
            "tags": self.build_tags(),
        }
    
    def get_cuboid_creator(self):
        return CuboidCreator(self.object_classes_map)




from abc import ABC
from typing import Optional
from uuid import uuid4

from ..common.label_creator import LabelCreator
from ..common.utils import set_properties, unique_string_builder, is_sublist
from .label import Cuboid
from ..common.label import ClassType


class CuboidCreator(LabelCreator):
    def __init__(self, object_classes_map):
        super().__init__(object_classes_map)
        self.label = Cuboid()

    def get_default_cuboid_coord(self):
        return {
            "position": {
                "x": 0.0,
                "y": 0.0,
                "z": 0.0
            },
            "rotation_quaternion": {
                "x": 0.0,
                "y": 0.0,
                "z": 0.0,
                "w": 1.0
            },
            "size": {
                "x": 1.0,
                "y": 1.0,
                "z": 1.0
            }
        }

    def create(
        self,
        class_name: str,
        frames: list,
        meta: dict = {},
        properties: list = [],
        id: str = str(uuid4()),
        tracking_id: Optional[int] = None,
    ):
        assert unique_string_builder(
            class_name,
            ClassType.CUBOID,
        ) in self.object_classes_map.keys()
        coord = frames[0]["annotation"]["coord"]
        assert is_sublist(
            ["position", "rotation_quaternion", "size"],
            coord.keys(),
        )
        return Cuboid(
            id=id,
            class_name=class_name,
            class_id=self.object_classes_map[
                unique_string_builder(
                    class_name,
                    ClassType.CUBOID,
                )
            ]["id"],
            frames=frames,
            meta=meta if meta else self.get_meta(
                class_name, ClassType.CUBOID),
            properties_def=self.object_classes_map[
                unique_string_builder(
                    class_name,
                    ClassType.CUBOID,
                )
            ]["properties"],
            properties=properties,
            tracking_id=tracking_id,
        )
    
    def from_dict(self, cuboid):
        class_name = cuboid["class_name"]
        assert unique_string_builder(
            class_name,
            ClassType.CUBOID,
        ) in self.object_classes_map.keys()
        frames = cuboid["frames"]

        return Cuboid(
            id=cuboid["id"] if "id" in cuboid else str(uuid4()),
            class_name=class_name,
            class_id=self.object_classes_map[
                unique_string_builder(
                    class_name,
                    ClassType.CUBOID,
                )
            ]["id"],
            frames=frames,
            properties_def=self.object_classes_map[
                unique_string_builder(
                    class_name,
                    ClassType.CUBOID,
                )
            ]["properties"],
            properties=cuboid["properties"] if "properties" in cuboid else [],
            tracking_id=cuboid["tracking_id"] if "tracking_id" in cuboid else 0,
        )

    def set_class(self, class_name):
        class_id = self.object_classes_map[
            unique_string_builder(
                class_name,
                ClassType.CUBOID,
            )
        ].get("id")

        if class_id is None:
            raise ValueError(f"Class name {class_name} not found in object classes map")
        
        self.label.class_name = class_name
        self.label.class_id = class_id

    def set_tracking_id(self, tracking_id):
        self.label.tracking_id = tracking_id
    # if frame information exists, update the frame information
    # else, add the frame information
    def set_annotation(self, frame_num, position=None, rotation_quaternion=None, size=None, frame_properties=None):
        target_frame = None
        for frame_info in self.label.frames:
            if frame_info["num"] == frame_num:
                target_frame = frame_info
                break
        
        if target_frame is None: 
            target_frame = {
                "num": frame_num,
                "annotation": {
                    "coord": self.get_default_cuboid_coord()
                },
                "properties": []
            }
            self.label.frames.append(target_frame)


        if position is not None:
            target_frame["annotation"]["coord"]["position"] = position
        if rotation_quaternion is not None:
            target_frame["annotation"]["coord"]["rotation_quaternion"] = rotation_quaternion
        if size is not None:
            target_frame["annotation"]["coord"]["size"] = size
        if frame_properties is not None:
            target_frame["properties"] = frame_properties

    def set_properties(self, properties, frame_num=None):
        if frame_num is None:
            self.label.properties = properties
        else:
            if self.label.frames.get(frame_num) is None:
                raise ValueError(f"Frame number {frame_num} not found in cuboid")
            else:
                self.label.frames[frame_num]["properties"] = properties

class PointcloudsAnnotationBuilder(ABC):
    def __init__(self, object_classes_map):
        self.object_classes_map = object_classes_map
    
    def get_cuboid_creator(self):
        return CuboidCreator(self.object_classes_map)

    def from_dict(self, label):
        creators = [
            CuboidCreator(self.object_classes_map),
        ]
        for creator in creators:
            try:
                return creator.from_dict(label)
            except:
                continue
        return None

